@@ -120,8 +120,13 @@ const double cm=inches/2.54;
 const double tex2ps=72.0/72.27;
 const double ps2tex=1.0/tex2ps;
 
+const string AsyGLVersion="1.00";
+const string AsyGL="webgl/asygl-"+ AsyGLVersion+".js";
+const string WebGLheader="webgl/WebGLheader.html";
+const string WebGLfooter="webgl/WebGLfooter.html";
 }
 
+
 extern const char *REVISION;
 
 #endif
